//------------------------------------------------------------------------------
//	$Id: bingo.js,v 1.4 2022/03/26 18:20:23 wolf Exp wolf $
//------------------------------------------------------------------------------

/*jsl:option explicit*/
/*jsl:import afm-helvetica.js*/
/*jsl:import pdfmaker.js*/

'use strict';

window.onload = function() {

	var RCS = '$Id: bingo.js,v 1.4 2022/03/26 18:20:23 wolf Exp wolf $';

	//----------------------------------------------------------------------
	//	create PDF
	//----------------------------------------------------------------------

	function createPDF (form, params) {

		var pdf = new PDFMAKER();

		pdf.setAuthor  (RCS.split('\x20').slice(1,4).join('\x20'));
		pdf.setCreator (navigator.userAgent + ' (' +
				form.getAttribute('data-pdf-source') + ')');
		pdf.setTitle   (form.getAttribute('data-pdf-title') || 'Bingo!');
		pdf.setSubject (form.getAttribute('data-pdf-subject'));

		pdf.setMediaSize (params.pageWidth, params.pageHeight);
		pdf.setMargins (params.marginLeft, params.marginBottom,
				params.marginRight, params.marginTop);

		//--------------------------------------------------------------
		//	coupon sizes and positions
		//--------------------------------------------------------------

		var nCols = params.cardsPerRow || 1;
		var nRows = params.cardsPerCol || 1;

		var gridSize = Math.min (
			(pdf.contentWidth  + params.cardSpacing) / nCols,
			(pdf.contentHeight + params.cardSpacing) / nRows);

		var payloadWidth = gridSize * nCols - params.cardSpacing;
		var payloadHeight= gridSize * nRows - params.cardSpacing;

		var xOffset = (pdf.contentWidth  - payloadWidth ) / 2;
		var yOffset = (pdf.contentHeight - payloadHeight) / 2;

		var cardSize   = gridSize - params.cardSpacing;
		var cellSize   = cardSize / params.cardFormat;

		//--------------------------------------------------------------
		//	style
		//--------------------------------------------------------------

		var style = {
			textColor: pdf.BLACK,
			borderColor: pdf.BLACK,
			borderWidth: 0,
			padding: 1,
			width: cellSize,
			height: cellSize,
			lineHeight: params.lineHeight || 1.2,
			lrAlign: 'c',
			tbAlign: 'm'
		};

		//--------------------------------------------------------------
		//	font
		//--------------------------------------------------------------

		var cellFont = pdf.getFont(params.fontName || 'Helvetica');

		var fontSize = (style.height - 2 * (style.padding || 0) ) /
				style.lineHeight / (params.linesPerCell || 6);

		//--------------------------------------------------------------
		//	create card frame
		//--------------------------------------------------------------

		var cardFrame = pdf.createBox({
			width:  cardSize,
			height: cardSize,
			borderWidth: 1,
			borderColor: pdf.BLACK
		}).getResult();

		//--------------------------------------------------------------
		//	create text boxes
		//--------------------------------------------------------------

		var textBoxes = [];

		for (var t=0; t<params.texts.length; t++) {

			var text = params.texts[t].
					replace(/[|]/g,'\xad').replace(/¦/g,'\n');

			var textBox = pdf.createTextBox (cellFont, fontSize, style);
			textBox.writeln (text);

			var result = textBox.getResult();

			textBoxes.push(result);
		}

		//--------------------------------------------------------------
		//	create pages
		//--------------------------------------------------------------

		for (var p=0; p<params.nPages; p++) {

			var page   = pdf.createPage();
			var writer = pdf.createWriterForPage(page);

			page._y -= yOffset;
			page._x += xOffset;

			for (var row=0; row<nRows; row++) {

				var cardY = page._y - row * gridSize;

				for (var col=0; col<nCols; col++) {

					//--------------------------------------
					//	create card frame
					//--------------------------------------

					var cardX = page._x + col * gridSize;

					writer ('q');
					writer ('1 0 0 1 ' + cardX + ' ' + cardY + ' cm');
					writer (cardFrame);

					//--------------------------------------
					//	create card cells
					//--------------------------------------

					var texts = [];

					for (var y=0; y<params.cardFormat; y++) {
						for (var x=0; x<params.cardFormat; x++) {

							//----------------------
							// create text cell
							//----------------------

							if (!texts.length) texts=textBoxes.slice();

							textBox = texts.splice(Math.random()*texts.length, 1)[0];

							writer ('q');
							writer ('1 0 0 1 ' + x*cellSize + ' ' + -y*cellSize + ' cm');
							writer (textBox);
							writer ('Q');
						}
					}

					writer ('Q');
				}
			}
		}

		return pdf;
	}

	//----------------------------------------------------------------------
	//	download PDF
	//----------------------------------------------------------------------

	function downloadPDF (form, params, pdf, aElement ) {

		//--------------------------------------------------------------
		//	create data:-link for pdf
		//--------------------------------------------------------------

		var mimetype= 'application/pdf';
		var pdfText = pdf.toString();
		var dataURL = 'data:'+mimetype+';base64,'+ window.btoa(pdfText);

		//--------------------------------------------------------------
		//	create download name
		//--------------------------------------------------------------

		function dd (n) { return n<10? '0' + n : n; }

		var date    = new Date();
		var stamp   = date.getFullYear() + '-' +
				dd(date.getMonth()+1) + '-' +
				dd(date.getDate()) + '-' +
				dd(date.getHours()) + '-' +
				dd(date.getMinutes()) + '-' +
				dd(date.getSeconds());
				
		var downloadName = (form.getAttribute('data-pdf-filename') ||
					'bingo') + '-' + stamp + '.pdf';

		//--------------------------------------------------------------
		//	confirm
		//--------------------------------------------------------------

		var message = form.getAttribute('data-pdf-confirm');
		if (message && !params.confirmed) {

			if (!window.confirm(message)) return;
			params.confirmed = true;
		}

		//--------------------------------------------------------------
		//	create and click link
		//--------------------------------------------------------------

		aElement.setAttribute('download', downloadName);
		aElement.setAttribute('type', mimetype);
		aElement.setAttribute('href', dataURL);
		aElement.style.display = '';
	}

	//----------------------------------------------------------------------
	//	set texts and remove unusable card formats
	//----------------------------------------------------------------------

	function setTexts (form, params, texts) {

		params.texts = texts;

		var cardFormatElement = form.elements['card-format'];
		if (!cardFormatElement) return;

		for (var format = cardFormatElement.firstElementChild; format;
				format = format.nextElementSibling) {

			format.disabled = 
				format.value*format.value > texts.length;
		}
	}

	//----------------------------------------------------------------------
	//	create XMLHttpRequest
	//----------------------------------------------------------------------

	function createXMLHttpRequest() {

		// Mozilla, Opera, Safari sowie Internet Explorer (ab v7)
		try {
			return new XMLHttpRequest();
		} catch(e) {}

		// MS Internet Explorer (ab v6)
		try {
			return ActiveXObject("Microsoft.XMLHTTP");
		} catch(e) {}

		// MS Internet Explorer (ab v5)
		try {
			return ActiveXObject("Msxml2.XMLHTTP");
		} catch(e) {}

		return null;
	}

	//----------------------------------------------------------------------
	//	read texts from external file using XMLHttpRequest
	//----------------------------------------------------------------------

	function readTexts (form, params, textURL) {

		var xhr = createXMLHttpRequest();
		if (!xhr) return;

		//--------------------------------------------------------------
		//	process response
		//--------------------------------------------------------------

		xhr.onload = function(evt) {

			if (this.status < 200 || this.status >=300) {

				window.alert ('Fetching "' + textURL +
					'" failed.\nStatus: ' +this.status);
				return;
			}

			var texts = [];

			var lines = this.responseText.replace(/\s+$/,'').split(/\n/);

			while (lines.length) {

				//----------------------------------------------
				//	empty line
				//----------------------------------------------

				var match = lines.shift().match(/^\s*(.*\S)/);
				if (!match) continue;
				var line = match[1];

				//----------------------------------------------
				//	text line
				//----------------------------------------------

				if (line.substr(0,2) != '#\x20') {

					texts.push(line);
					continue;
				}

				//----------------------------------------------
				//	comment line
				//----------------------------------------------

				match = line.match(/^#\s*TITLE\s+(.*\S)$/);
				if (match) {

					form.setAttribute('data-pdf-title', match[1]);
					continue;
				}

				match = line.match(/^#\s*PDF_FILENAME\s+(\w+(?:[\/.-]\w+)*)$/);
				if (match) {

					form.setAttribute('data-pdf-filename', match[1]);
					continue;
				}

				match = line.match(/^#\s*FONTNAME\s+([A-Z][a-z]+(?:-?[A-Z][a-z]+)*)$/);
				if (match) {

					params.fontName = match[1];
					continue;
				}

				match = line.match(/^#\s*LINES_PER_CELL\s+(\d+\.?\d*)$/);
				if (match) {

					params.linesPerCell = parseFloat(match[1]);
					continue;
				}

				window.console.log ('Invalid line in ' + textURL + ': ' + line);
			}

			setTexts (form, params, texts.sort());
		};

		xhr.open('GET', textURL);
		xhr.send(null);
	}

	//----------------------------------------------------------------------
	//	get texts from data-attributes or external URL
	//----------------------------------------------------------------------

	function getTexts (form, params) {

		//--------------------------------------------------------------
		//	if URL available, use XMLHttpRequest
		//--------------------------------------------------------------

		var textURL = form.getAttribute('data-bingo-text-url');
		if (textURL) {

			readTexts(form, params, textURL);
			return;
		}

		//--------------------------------------------------------------
		//	get texts from attributes
		//--------------------------------------------------------------

		var texts = [];

		for (var a=0; a<form.attributes.length; a++) {

			var attr = form.attributes[a];

			if (!attr.name.match(/^data-bingo-text-\d+$/)) continue;
			texts.push(attr.value);
		}

		//--------------------------------------------------------------
		//	if no texts found, create default bingo numbers
		//--------------------------------------------------------------

		if (texts.length) {

			params.linesPerCell =
			parseFloat(form.getAttribute('data-bingo-lines-per-cell'));
		} else {
			for (var b=1; b<=25; b++) { texts.push(b.toString()); }
			params.linesPerCell = 1;
		}

		setTexts (form, params, texts.sort());
	}

	//----------------------------------------------------------------------
	//	add or remove entries to cards per page select
	//----------------------------------------------------------------------

	function updateCardsPerPage (form, params) {

		var cpp = form.elements['cards-per-page'];
		if (!cpp || cpp.tagName != 'SELECT') return;

		var width  = params.pageWidth  - params.marginLeft- params.marginRight  + params.cardSpacing;
		var height = params.pageHeight - params.marginTop - params.marginBottom + params.cardSpacing;
		var limit  = params.cardFormat * params.cellSize  + params.cardSpacing;
		if (isNaN(width) || isNaN(height) || isNaN(limit)) return;

		var oldNCards = cpp.value.split(':')[0] * cpp.value.split(':')[1];

		var x = 1;
		var y = 1;
		var s = Math.min (width/x, height/y);
		var index;

		for (index=0; x==1 && y==1 || s >= limit; index++) {

			var x_y = x + ':' + y;
			var xy  = x * y;

			if (index < cpp.children.length && cpp.children[index].value != x_y) {

				while (index < cpp.children.length) {

					cpp.removeChild(cpp.lastElementChild);
				}
			}

			var option = cpp.children[index];

			if (index >= cpp.children.length) {

				option = document.createElement('option');
				option.value = x_y;
				option.appendChild(document.createTextNode(xy));
				cpp.appendChild(option);
			}

			if (xy <= oldNCards) cpp.value = option.value;
	
			var sx = width  / (x+1);
			var sy = height / (y+1);

			if (sx >= sy) x++;
			if (sx <= sy) y++;

			s = Math.max(sx, sy);
		}

		while (index < cpp.children.length) {

			cpp.removeChild(cpp.lastElementChild);
		}
	}

	//----------------------------------------------------------------------
	//	conversions
	//----------------------------------------------------------------------

	function ptFromMM(mm) { return parseFloat(mm) / 25.4 * 72; }

	//----------------------------------------------------------------------
	//	change numeric parameters
	//----------------------------------------------------------------------

	function onchange (params) {

		switch (this.name) {

		case 'theme':

			params.texts = null;
			if (this.value) readTexts (form, params, this.value);
			return;

		case 'page-size-mm':

			params.pageWidth = ptFromMM(this.value.split(':')[0]);
			params.pageHeight= ptFromMM(this.value.split(':')[1]);
			break;

		case 'page-width-mm':

			params.pageWidth = ptFromMM(this.value);
			break;

		case 'page-height-mm':

			params.pageHeight = ptFromMM(this.value);
			break;

		case 'margin-top-bottom-mm':

			params.marginTop    = ptFromMM(this.value);
			params.marginBottom = ptFromMM(this.value);
			break;

		case 'margin-top-mm':

			params.marginTop = ptFromMM(this.value);
			break;

		case 'margin-bottom-mm':

			params.marginBottom = ptFromMM(this.value);
			break;

		case 'margin-left-right-mm':

			params.marginLeft  = ptFromMM(this.value);
			params.marginRight = ptFromMM(this.value);
			break;

		case 'margin-left-mm':

			params.marginLeft = ptFromMM(this.value);
			break;

		case 'margin-right-mm':

			params.marginRight = ptFromMM(this.value);
			break;

		case 'card-spacing-mm':

			params.cardSpacing = ptFromMM(this.value);
			break;

		case 'cell-size-mm':

			params.cellSize = ptFromMM(this.value);
			break;

		case 'card-format':

			params.cardFormat = parseInt(this.value, 10);
			break;

		case 'cards-per-page':

			params.cardsPerRow = parseInt(this.value.split(':')[0], 10);
			params.cardsPerCol = parseInt(this.value.split(':')[1], 10);
			return;

		case 'font-name-and-lines-per-cell':

			params.fontName = this.value.split(':')[0].replace(/\s+/g,'');
			params.linesPerCell = parseFloat(this.value.split(':')[1]);
			return;

		case 'font-name':

			params.fontName = this.value.replace(/\s+/g,'');
			return;

		case 'lines-per-cell':

			params.linesPerCell = parseFloat(this.value);
			return;

		case 'line-height':

			params.lineHeight = parseFloat(this.value);
			return;

		case 'n-pages':

			params.nPages = parseInt(this.value, 10);
			return;

		default:
			window.console.log ('UNKNOWN form element name: ' + this.name);
			return;
		}

		updateCardsPerPage (this.form, params);

	}

	//----------------------------------------------------------------------
	//	create a single bingo
	//----------------------------------------------------------------------

	function installBingo (form) {

		//--------------------------------------------------------------
		//	configure bingo
		//--------------------------------------------------------------

		var params = {
			nPages:       1,
			cardFormat:   5,
			pageHeight:   ptFromMM(297.3),
			pageWidth:    ptFromMM(210.2),
			marginTop:    ptFromMM(25),
			marginBottom: ptFromMM(25),
			marginLeft:   ptFromMM(13),
			marginRight:  ptFromMM(13),
			cardSpacing:  ptFromMM(8),
			cellSize:     ptFromMM(10)
		};

		getTexts (form, params);

		//--------------------------------------------------------------
		//	create download link a element
		//--------------------------------------------------------------

		var button       = form.elements[ 'create-pdf' ];
		var downloadLink = document.createElement( 'a' );

		downloadLink.style.display = 'none';

		downloadLink.appendChild( document.createTextNode(
			button.getAttribute('download-label') ||
				'download PDF' ));

		//--------------------------------------------------------------
		//	install download link a element
		//--------------------------------------------------------------

		if( button.nextSibling ){

			button.parentElement.insertBefore(
				downloadLink, button.nextSibling );
		} else {
			button.parentElement.appendChild( downloadLink );
		}

		//--------------------------------------------------------------
		//	process form elements
		//--------------------------------------------------------------

		for (var e=0; e<form.elements.length; e++) {

			var element = form.elements[e];

			switch(element.tagName) {

			case 'INPUT':
			case 'SELECT':

				element.onchange = function () {

					downloadLink.style.display = 'none';
					onchange.apply(this, [params]);
				};

				element.onchange(null);
				break;

			case 'BUTTON':

				if (element.name=='create-pdf') {

					element.onclick = function() {

						downloadPDF (form, params,
							createPDF(form, params),
							downloadLink);
					};
					break;
				}
				break;

			default:
				break;
			}
		}

		updateCardsPerPage (form, params);
	}

	//----------------------------------------------------------------------
	//	lookup bingo forms
	//----------------------------------------------------------------------

	for (var f=0; f<document.forms.length; f++) {

		var form = document.forms[f];

		if (form.elements['create-pdf']) installBingo(form);
	}
};

//------------------------------------------------------------------------------
//	$Id: bingo.js,v 1.4 2022/03/26 18:20:23 wolf Exp wolf $
//------------------------------------------------------------------------------
